/* GECKO3COM
 *
 * Copyright (C) 2009 by
 *   ___    ____  _   _
 *  (  _`\ (  __)( ) ( )   
 *  | (_) )| (_  | |_| |   Berne University of Applied Sciences
 *  |  _ <'|  _) |  _  |   School of Engineering and
 *  | (_) )| |   | | | |   Information Technology
 *  (____/'(_)   (_) (_)
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*********************************************************************/
/** \file     scpi_parser.re
 *********************************************************************
 * \brief     Parser for the SCPI commands. Excecutes the necessary 
 *            Function.
 *
 *            The SCPI Parser is developed as a set of regular 
 *            expressions and blocks of C code to excecute. \n
 *            We use the opensource programm re2c (http://re2c.org/) to
 *            generate the parser. With solution we have an easy to 
 *            maintain and extend parser definitions without the hassle 
 *            of optimizing the parser sourcecode by hand. \n \n
 *            Don't forget to use the "-s" option with re2c to generate
 *            nested ifs for some switches because sdcc hates long 
 *            switch case structures.
 *
 * \todo      use newer re2c to generate state event diagramm to include here
 *
 * \author    Christoph Zimmermann bfh.ch
 * \date      2009-02-04
 *
*/

#include <stdint.h>
#include <ctype.h>
#include <string.h>

#include "scpi_parser.h"
#include "usb_tmc.h"
#include "usb_descriptors.h"
#include "firmware_version.h"
#include "debugprint.h"


#define SCPI_VERSION "1999.0\n" /**< Version of the SCPI standard this device complies to */

#define	YYCTYPE		unsigned char
#define	YYCURSOR	s->cur
#define	YYLIMIT		s->lim
#define	YYMARKER	s->ptr
#define	YYFILL(n)	{YYCURSOR = fill(s, buffer, n);}
/*!max:re2c */          /* this will define YYMAXFILL */

/** helper function to copy unicode 16 chars from the usb descriptor to an 
 * ascii string to form the *idn? response 
 * 
 * \param[in] pointer to string to fill
 * \param[in] pointer to usb string descriptor to read from
 * \return pointer that points to the last character in the string (one before \0)
*/
char* descr2string(char *target, char *descriptor) {
  uint8_t i;
  /* the first byte in the descriptor is the length, second byte is the type */
  for (i = 0; i < (uint8_t)descriptor[0]-2; i=i+2){
    target[i >> 1] = descriptor[i+2];
  }
  i = i >> 1;
  target[i+1] = '\0';

  return &target[i];
}

/** internal helper function to fill the buffer with n characters. 
 *  we convert any character to lower case because SCPI is case insensitive 
 *
 *  \param[in]  *s pointer to a Scanner struct. contains all needed pointers for the SCPI parser
 *  \param[out] *buffer pointer to the buffer to fill
 *  \param[in]  n number of bytes to read and fill into the buffer
 *  \return pointer to the next fresh character to parse
*/
unsigned char* fill(Scanner *s, unsigned char *buffer, uint8_t n){
  uint8_t i = 0;

  for(i;i<n;i++) {
    buffer[i] = tolower(*(s->source));
    s->source++;
  }

  s->cur = buffer;
  s->lim = &buffer[n];

  return s->cur;    
}

/** SCPI command parser */
int8_t scpi_scan(Scanner *s, TMC_Response_Queue *queue){

  unsigned char buffer[YYMAXFILL];
  unsigned char *i;

  /** \cond SCANNER  
   * this is a ugly hack to avoid that the regular expressions are included in 
   * the doxygen documentation
  */

  /*!re2c	

  /* this set of regular expressions are for the mandatory IEEE488 commands */

  "*cls"				{ /* clear status command */
  					  ieee488_status.EventStatusRegister = 0;
					  ieee488_status.StatusByteRegister = 0;
					  usb_tmc_state = TMC_STATE_IDLE;
					  return 1;  
					}

  "*ese "				{ /* standard event status enable command */
  					  ieee488_status.EventStatusEnable = *++YYCURSOR;
					  return 1;  
					}

  "*ese?"				{ /* standard event status enable query */
  					  queue->buf[0] = ieee488_status.EventStatusEnable;
					  queue->length = 1;
					  IEEE488_set_mav();
					  usb_tmc_state = TMC_STATE_IDLE;
					  return 1;  
					}

  "*esr?"				{ /* standard event status register query */
  					  queue->buf[0] = ieee488_status.EventStatusRegister;
					  queue->length = 1;
					  IEEE488_set_mav();
					  ieee488_status.EventStatusRegister = 0;
					  usb_tmc_state = TMC_STATE_IDLE;
					  return 1;  
					}

  "*idn?"				{ /* identification query */
					  i = descr2string((char*)queue->buf, string_descriptors[(uint8_t)full_speed_device_descr[14]]);
					  *i = ',';
					  i = descr2string(&i[1], string_descriptors[(uint8_t)full_speed_device_descr[15]]);
					  *i = ',';
					  i = descr2string(&i[1], string_descriptors[(uint8_t)full_speed_device_descr[16]]);
					  
					  strcat((char*)queue->buf, ",");
					  strcat((char*)queue->buf, FIRMWARE_VERSION);
					  
					  queue->length = strlen((char*)queue->buf);

					  usb_tmc_state = TMC_STATE_IDLE;
					  IEEE488_set_mav();
					  return 1; 
					}

  "*opc"				{ /* operation complete command */
  					  ieee488_status.OPC_Received = 1;
					  usb_tmc_state = TMC_STATE_IDLE;
					  return 1;  
					}

  "*opc?"				{ /* operation complete query */
  					  queue->buf[0] = 1;
					  queue->length = 1;
					  usb_tmc_state = TMC_STATE_IDLE;
					  IEEE488_set_mav();
					  return 1;  
					}

  "*rst"				{ /* reset command */
  					  usb_tmc_state = TMC_STATE_IDLE;
					  return 1;  
					}

  "*sre "				{ /* service request enable command */
  					  ieee488_status.ServiceRequestEnable = *++YYCURSOR;
					  return 1;  
					}

  "*sre?"				{ /* service request enable query */
  					  queue->buf[0] = ieee488_status.ServiceRequestEnable;
					  queue->length = 1;
					  usb_tmc_state = TMC_STATE_IDLE;
					  return 1;  
					}

  "*stb?"				{ /* read status byte query */
  					  queue->buf[0] = IEEE488_status_query(&ieee488_status);
					  queue->length = 1;
					  usb_tmc_state = TMC_STATE_IDLE;
					  IEEE488_set_mav();
					  return 1;  
					}


  "*wai"				{ /* wait-to-continue command */
  					  /* we excecute only sequential commands, so we are always finished */
					  return 1;  
					}



  /* this set of regular expressions are for the mandatory SCPI 99 commands. 
   * many are missing because we have not enought memory 
  */

  "syst:err?" 	 	 	     	{ /* gets an error message if ther is one */
				  	  if(ieee488_status.EventStatusRegister & bmCOMMAND_ERROR){
					    strcpy((char*)queue->buf, "-100, \"Command error\"\n");
					    queue->length = 22;
					    ieee488_status.EventStatusRegister |= ~bmCOMMAND_ERROR;
					  }
					  else {
					    strcpy((char*)queue->buf, "0, \"No error\"\n");
					    queue->length = 14;
					  }
					  usb_tmc_state = TMC_STATE_IDLE;
					  IEEE488_set_mav();
					  return 1;  
					}

  "syst:vers?"				{ /* returns the firmware version number */
  					  strcpy((char*)queue->buf, SCPI_VERSION);
					  queue->length = 6;
					  usb_tmc_state = TMC_STATE_IDLE;
					  IEEE488_set_mav();
					  return 1;  
					}



  /* this set of regular expressions are for the device functions */


  "fpga:type?" 	       		       	{ /* ask if there is a message to read available */
				  	  print_info("fpga:type?\n"); 
					  return 1;  
					}

  "fpga:id?"				{ /* reads the message from the message buffer */
				  	  print_info("fpga:id?\n"); 
				  	  return 1; 
					}

  "fpga:done?"				{ /* reads the message from the message buffer */
				  	  print_info("fpga:done?\n"); 
				  	  return 1; 
					}

  "fpga:conf "				{ /* configure the fpga */

  	     	     			}


/* matches all. when no command is parsed return an error */
  [^] | "\n"				{ return 0; }

*/

/** \endcond */
}
